import gettext as gettext_module
import sys
from gettext import NullTranslations
from typing import Any, Callable, Dict, Iterator, List, Optional, Pattern, Protocol, Tuple, TypeVar, Union

from django.http.request import HttpRequest

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

CONTEXT_SEPARATOR: Literal["\x04"]
accept_language_re: Pattern[str]
language_code_re: Pattern[str]
language_code_prefix_re: Pattern[str]

class _PluralCallable(Protocol):
    def __call__(self, __n: int) -> int: ...

def reset_cache(**kwargs: Any) -> None: ...

_KeyT = Union[str, Tuple[str, int]]
_Z = TypeVar("_Z")

class TranslationCatalog:
    _catalogs: List[Dict[_KeyT, str]]
    def __init__(self, trans: Optional[gettext_module.NullTranslations] = ...) -> None: ...
    def __getitem__(self, key: _KeyT) -> str: ...
    def __setitem__(self, key: _KeyT, value: str) -> None: ...
    def __contains__(self, key: _KeyT) -> bool: ...
    def items(self) -> Iterator[Tuple[_KeyT, str]]: ...
    def keys(self) -> Iterator[_KeyT]: ...
    def update(self, trans: gettext_module.NullTranslations) -> None: ...
    def get(self, key: _KeyT, default: _Z = ...) -> Union[str, _Z]: ...
    def plural(self, msgid: str, num: int) -> str: ...

class DjangoTranslation(gettext_module.GNUTranslations):
    domain: str = ...
    plural: _PluralCallable = ...
    def __init__(self, language: str, domain: Optional[str] = ..., localedirs: Optional[List[str]] = ...) -> None: ...
    def merge(self, other: NullTranslations) -> None: ...
    def language(self) -> str: ...
    def to_language(self) -> str: ...
    def ngettext(self, msgid1: str, msgid2: str, n: int) -> str: ...

def translation(language: str) -> DjangoTranslation: ...
def activate(language: str) -> None: ...
def deactivate() -> None: ...
def deactivate_all() -> None: ...
def get_language() -> str: ...
def get_language_bidi() -> bool: ...
def catalog() -> DjangoTranslation: ...
def gettext(message: str) -> str: ...
def pgettext(context: str, message: str) -> str: ...
def gettext_noop(message: str) -> str: ...
def do_ntranslate(singular: str, plural: str, number: float, translation_function: str) -> str: ...
def ngettext(singular: str, plural: str, number: float) -> str: ...
def npgettext(context: str, singular: str, plural: str, number: int) -> str: ...
def all_locale_paths() -> List[str]: ...
def check_for_language(lang_code: Optional[str]) -> bool: ...
def get_languages() -> Dict[str, str]: ...
def get_supported_language_variant(lang_code: Optional[str], strict: bool = ...) -> str: ...
def get_language_from_path(path: str, strict: bool = ...) -> Optional[str]: ...
def get_language_from_request(request: HttpRequest, check_path: bool = ...) -> str: ...
def parse_accept_lang_header(lang_string: str) -> Tuple[Tuple[str, float], ...]: ...
